@@ -4,7 +4,7 @@ import re
 from collections import defaultdict
 from json import load, dump
 from multiprocessing.pool import ThreadPool
-from socket import gethostbyname
+from socket import gethostbyname, gaierror
 
 import boto3
 from pkg_resources import resource_stream, resource_filename
@@ -17,15 +17,18 @@ cache = AppCache('aws_list_all')
 
 VERBS_LISTINGS = ['Describe', 'Get', 'List']
 
-SERVICE_BLACKLIST = [
+SERVICE_IGNORE_LIST = [
     'alexaforbusiness',  # TODO: Mostly organization-specific calls and would need to be queried differently
     'apigatewaymanagementapi',  # This API allows management of deployed APIs, and requires an endpoint per API.
+    'backupstorage',  # This seems to be an API centerered around Jobs, no listings possible
     'cloudsearchdomain',  # Domain-specific endpoint required
     'kinesis-video-archived-media',  # API operating on stream-specific endpoints
     'kinesis-video-media',  # API operating on stream-specific endpoints
+    'macie',  # This service has been deprecated and turned off
     'managedblockchain',  # TODO: Unclear, does not have a region
     'mediastore-data',  # Mediastore Container-specific endpoint required
     's3control',  # TODO: Account-ID specific endpoint required
+    'worklink',  # Seems to have no API?
 ]
 
 DEPRECATED_OR_DISALLOWED = {
@@ -112,7 +115,9 @@ AWS_RESOURCE_QUERIES = {
         'DescribeSpotPriceHistory',
         'DescribeVpcClassicLinkDnsSupport',
         'DescribeVpcEndpointServices',
+        'GetVpnConnectionDeviceTypes',
     ],
+    'eks': ['DescribeAddonVersions'],
     'elasticache': ['DescribeCacheParameterGroups', 'DescribeCacheEngineVersions', 'DescribeServiceUpdates'],
     'elasticbeanstalk': [
         'ListAvailableSolutionStacks',
@@ -123,6 +128,7 @@ AWS_RESOURCE_QUERIES = {
     'elb': ['DescribeLoadBalancerPolicyTypes', 'DescribeLoadBalancerPolicies'],
     'elbv2': ['DescribeSSLPolicies'],
     'es': ['DescribeReservedElasticsearchInstanceOfferings', 'GetCompatibleElasticsearchVersions'],
+    'fis': ['ListActions'],
     'groundstation': ['ListGroundStations'],
     'inspector': ['ListRulesPackages'],
     'kafka': ['GetCompatibleKafkaVersions', 'ListKafkaVersions'],
@@ -136,6 +142,8 @@ AWS_RESOURCE_QUERIES = {
     'mobile': ['ListBundles'],
     'mq': ['DescribeBrokerInstanceOptions', 'DescribeBrokerEngineTypes'],
     'neptune': ['DescribeDBEngineVersions', 'DescribeEventCategories'],
+    'opensearch': ['DescribeReservedInstanceOfferings'],
+    'outposts': ['ListCatalogItems'],
     'personalize': ['ListRecipes'],
     'pricing': ['DescribeServices'],
     'polly': ['DescribeVoices'],
@@ -148,14 +156,20 @@ AWS_RESOURCE_QUERIES = {
         'DescribeEventCategories',
         'DescribeClusterTracks',
     ],
+    'resiliencehub': ['ListSuggestedResiliencyPolicies'],
+    'resource-explorer-2': ['ListSupportedResourceTypes'],
     'route53': ['GetCheckerIpRanges', 'ListGeoLocations'],
+    'route53domains': ['ListPrices'],
     'savingsplans': ['DescribeSavingsPlansOfferingRates', 'DescribeSavingsPlansOfferings'],
-    'securityhub': ['DescribeStandards'],
+    'securityhub': ['DescribeStandards', 'DescribeProducts', 'GetEnabledStandards', 'ListEnabledProductsForImport'],
     'service-quotas': ['ListServices'],
     'signer': ['ListSigningPlatforms'],
     'ssm': ['DescribeAvailablePatches', 'GetInventorySchema'],
     'synthetics': ['DescribeRuntimeVersions'],
+    'timestream-query': ['DescribeEndpoints'],
+    'timestream-write': ['DescribeEndpoints'],
     'transfer': ['ListSecurityPolicies'],
+    'translate': ['ListLanguages'],
     'xray': ['GetSamplingRules'],
 }
 
@@ -165,10 +179,16 @@ AWS_RESOURCE_QUERIES = {
 # still be used later for change tracking, e.g. tracking account limits over time with DescribeAccountLimits.
 NOT_RESOURCE_DESCRIPTIONS = {
     'apigateway': ['GetAccount'],
+    'account': ['GetContactInformation'],
+    'acm': ['GetAccountConfiguration'],
+    'auditmanager': ['GetAccountStatus'],
     'autoscaling': ['DescribeAccountLimits'],
     'alexaforbusiness': ['GetInvitationConfiguration'],
+    'appflow': ['DescribeConnectors', 'ListConnectorEntities', 'ListConnectors'],
     'athena': ['ListQueryExecutions'],
-    'chime': ['GetGlobalSettings'],
+    'backup': ['DescribeGlobalSettings', 'DescribeRegionSettings'],
+    'chime': ['GetGlobalSettings', 'GetMessagingSessionEndpoint'],
+    'chime-sdk-messaging': ['GetMessagingSessionEndpoint'],
     'cloudformation': ['DescribeAccountLimits'],
     'cloudwatch': ['DescribeAlarmHistory'],
     'codebuild': ['ListBuilds'],
@@ -176,33 +196,68 @@ NOT_RESOURCE_DESCRIPTIONS = {
         'GetComplianceSummaryByResourceType', 'GetComplianceSummaryByConfigRule', 'DescribeComplianceByConfigRule',
         'DescribeComplianceByResource', 'DescribeConfigRuleEvaluationStatus', 'GetDiscoveredResourceCounts'
     ],
+    'compute-optimizer': ['GetEnrollmentStatus'],
     'dax': ['DescribeDefaultParameters', 'DescribeParameterGroups'],
     'devicefarm': ['GetAccountSettings', 'GetOfferingStatus'],
+    'devops-guru': [
+        'DescribeOrganizationHealth',
+        'DescribeAccountHealth',
+        'DescribeServiceIntegration',
+        'GetCostEstimation',
+    ],
+    'directconnect': ['DescribeCustomerMetadata'],
     'discovery': ['GetDiscoverySummary'],
     'dms': ['DescribeAccountAttributes', 'DescribeApplicableIndividualAssessments', 'DescribeEventCategories'],
     'docdb': ['DescribeEvents'],
     'ds': ['GetDirectoryLimits'],
     'dynamodb': ['DescribeLimits'],
     'ec2': [
-        'DescribeAccountAttributes', 'DescribeDhcpOptions', 'DescribeVpcClassicLink',
-        'DescribeVpcClassicLinkDnsSupport', 'DescribePrincipalIdFormat', 'GetEbsDefaultKmsKeyId',
-        'GetEbsEncryptionByDefault'
+        'DescribeAccountAttributes',
+        'DescribeDhcpOptions',
+        'DescribeInstanceEventNotificationAttributes',
+        'DescribePrincipalIdFormat',
+        'DescribeVpcClassicLink',
+        'DescribeVpcClassicLinkDnsSupport',
+        'GetEbsDefaultKmsKeyId',
+        'GetEbsEncryptionByDefault',
+        'GetSerialConsoleAccessStatus',
+        'InstanceTagAttribute',
     ],
-    'ecr': ['GetAuthorizationToken'],
+    'ecr': ['GetAuthorizationToken', 'DescribeRegistry', 'GetRegistryPolicy', 'GetRegistryScanningConfiguration'],
+    'ecr-public': ['GetAuthorizationToken', 'GetRegistryCatalogData'],
     'ecs': ['DescribeClusters'],  # This gives duplicates from ListClusters, and also includes deleted clusters
+    'efs': ['DescribeAccountPreferences', 'DescribeReplicationConfigurations'],
     'elastic-inference': ['DescribeAcceleratorTypes'],
     'elasticache': ['DescribeReservedCacheNodesOfferings'],
     'elasticbeanstalk': ['DescribeAccountAttributes', 'DescribeEvents'],
     'elb': ['DescribeAccountLimits'],
     'elbv2': ['DescribeAccountLimits'],
+    'emr': ['DescribeReleaseLabel', 'GetBlockPublicAccessConfiguration'],
     'es': ['ListElasticsearchVersions'],
     'events': ['DescribeEventBus'],
     'fms': ['GetAdminAccount', 'GetNotificationChannel'],
+    'frauddetector': ['GetKMSEncryptionKey'],
     'gamelift': ['DescribeEC2InstanceLimits', 'DescribeMatchmakingConfigurations', 'DescribeMatchmakingRuleSets'],
     'glue': ['GetCatalogImportStatus', 'GetDataCatalogEncryptionSettings'],
     'guardduty': ['GetInvitationsCount'],
+    'greengrassv2': ['GetServiceRoleForAccount'],
+    'healthlake': ['DescribeFHIRDatastore'],
     'iam': ['GetAccountPasswordPolicy', 'GetAccountSummary', 'GetUser', 'GetAccountAuthorizationDetails'],
+    'iotdeviceadvisor': ['GetEndpoint'],
+    'iotfleetwise': ['GetLoggingOptions', 'GetRegisterAccountStatus'],
+    'iotsitewise': [
+        'DescribeDefaultEncryptionConfiguration',
+        'DescribeLoggingOptions',
+        'DescribeStorageConfiguration',
+    ],
     'inspector': ['DescribeCrossAccountAccessRole'],
+    'inspector2': [
+        'DescribeOrganizationConfiguration',
+        'GetConfiguration',
+        'GetDelegatedAdminAccount',
+        'GetFindingsReportStatus',
+        'ListCoverageStatistics',
+    ],
     'iot': [
         'DescribeAccountAuditConfiguration',
         'DescribeEndpoint',
@@ -213,14 +268,26 @@ NOT_RESOURCE_DESCRIPTIONS = {
         'ListV2LoggingLevels',
     ],
     'iotevents': ['DescribeLoggingOptions'],
+    'iotwireless': ['GetEventConfigurationByResourceTypes', 'GetServiceEndpoint'],
     'iotthingsgraph': ['DescribeNamespace', 'GetNamespaceDeletionStatus'],
     'kinesis': ['DescribeLimits'],
     'lambda': ['GetAccountSettings'],
+    'lakeformation': ['GetDataLakeSettings'],
+    'macie2': [
+        'GetClassificationExportConfiguration',
+        'GetInvitationsCount',
+        'GetUsageStatistics',
+        'GetUsageTotals',
+    ],
+    'mediaconvert': ['GetPolicy'],
+    'migrationhubstrategy': ['GetPortfolioSummary'],
     'neptune': ['DescribeEvents'],
+    'networkmanager': ['ListOrganizationServiceAccessStatus'],
     'opsworks': ['DescribeMyUserProfile', 'DescribeUserProfiles', 'DescribeOperatingSystems'],
     'opsworkscm': ['DescribeAccountAttributes'],
     'organizations': ['DescribeOrganization'],
     'pinpoint-email': ['GetAccount', 'GetDeliverabilityDashboardOptions'],
+    'proton': ['GetAccountSettings'],
     'redshift': ['DescribeStorage', 'DescribeAccountAttributes'],
     'rds': [
         'DescribeAccountAttributes', 'DescribeDBEngineVersions', 'DescribeReservedDBInstancesOfferings',
@@ -229,15 +296,16 @@ NOT_RESOURCE_DESCRIPTIONS = {
     'resourcegroupstaggingapi': ['GetResources', 'GetTagKeys', 'DescribeReportCreation', 'GetComplianceSummary'],
     'route53': ['GetTrafficPolicyInstanceCount', 'GetHostedZoneCount', 'GetHealthCheckCount', 'GetGeoLocation'],
     'route53domains': ['ListOperations'],
-    'sagemaker': ['ListTrainingJobs'],
-    'securityhub': ['GetInvitationsCount'],
+    'route53resolver': ['ListResolverQueryLogConfigAssociations', 'ListResolverQueryLogConfigs'],
+    'sagemaker': ['ListTrainingJobs', 'GetSagemakerServicecatalogPortfolioStatus'],
+    'securityhub': ['GetInvitationsCount', 'DescribeHub', 'DescribeOrganizationConfiguration'],
     'servicediscovery': ['ListOperations'],
     'ses': ['GetSendQuota', 'GetAccountSendingEnabled'],
     'sesv2': ['GetAccount', 'GetDeliverabilityDashboardOptions'],
-    'shield': ['GetSubscriptionState'],
+    'shield': ['GetSubscriptionState', 'DescribeAttackStatistics'],
     'sms': ['GetServers'],
     'snowball': ['GetSnowballUsage'],
-    'sns': ['GetSMSAttributes', 'ListPhoneNumbersOptedOut'],
+    'sns': ['GetSMSAttributes', 'ListPhoneNumbersOptedOut', 'GetSMSSandboxAccountStatus'],
     'ssm': ['GetDefaultPatchBaseline'],
     'sts': ['GetSessionToken', 'GetCallerIdentity'],
     'waf': ['GetChangeToken'],
@@ -246,17 +314,36 @@ NOT_RESOURCE_DESCRIPTIONS = {
     'workspaces': ['DescribeAccount', 'DescribeAccountModifications'],
 }
 
+# TODO wafv2 ListWebAcls needs a scope parameter - this may not be caught here!
+
 PARAMETERS_REQUIRED = {
-    'appstream': ['DescribeUserStackAssociations'],
+    'appstream': ['DescribeUserStackAssociations', 'DescribeApplicationFleetAssociations'],
+    'application-insights': ['ListConfigurationHistory'],
     'batch': ['ListJobs'],
-    'cloudformation': ['GetTemplateSummary', 'DescribeStackResources', 'DescribeStackEvents', 'GetTemplate'],
+    'chime': ['ListChannelMembershipsForAppInstanceUser', 'ListChannelsModeratedByAppInstanceUser'],
+    'cloudformation': [
+        'DescribeStackEvents', 'DescribeStackResources', 'DescribeType', 'GetTemplate', 'GetTemplateSummary',
+        'ListTypeVersions'
+    ],
+    'cloudfront': [
+        'GetRealtimeLogConfig',
+        'ListDistributionsByRealtimeLogConfig',
+    ],
     'cloudhsm': ['DescribeHsm', 'DescribeLunaClient'],
     'cloudtrail': ['GetEventSelectors'],
+    'codebuild': ['ListBuildBatchesForProject'],
     'codecommit': ['GetBranch'],
     'codedeploy': ['GetDeploymentTarget', 'ListDeploymentTargets'],
     'cognito-idp': ['GetUser'],
+    'devops-guru': ['DescribeFeedback'],
     'directconnect': ['DescribeDirectConnectGatewayAssociations', 'DescribeDirectConnectGatewayAttachments'],
-    'ec2': ['DescribeSpotDatafeedSubscription', 'DescribeLaunchTemplateVersions'],
+    'dms': ['ListTagsForResource'],
+    'ec2': [
+        'DescribeLaunchTemplateVersions',
+        'DescribeSpotDatafeedSubscription',
+        'GetAssociatedEnclaveCertificateIamRoles',
+        'GetTransitGatewayMulticastDomainAssociations',
+    ],
     'ecs': ['ListContainerInstances', 'ListServices', 'ListTasks'],
     'efs': ['DescribeMountTargets'],
     'elasticache': ['ListAllowedNodeTypeModifications', 'DescribeCacheSecurityGroups'],
@@ -268,25 +355,52 @@ PARAMETERS_REQUIRED = {
     'elbv2': ['DescribeRules', 'DescribeListeners'],
     'gamelift': ['DescribeGameSessionDetails', 'DescribeGameSessions', 'DescribePlayerSessions'],
     'globalaccelerator': ['DescribeAcceleratorAttributes'],
-    'glue': ['GetDataflowGraph', 'GetResourcePolicy'],
+    'glue': ['GetDataflowGraph', 'GetResourcePolicy', 'GetSchemaVersion'],
     'health': [
         'DescribeEventTypes', 'DescribeEntityAggregates', 'DescribeEvents', 'DescribeEventsForOrganization',
         'DescribeHealthServiceStatusForOrganization'
     ],
-    'iot': ['GetLoggingOptions', 'GetEffectivePolicies', 'ListAuditFindings'],
+    'iot': ['GetLoggingOptions', 'GetEffectivePolicies', 'ListAuditFindings', 'ListDetectMitigationActionsExecutions'],
+    'iotsitewise': [
+        'DescribeTimeSeries',
+        'GetAssetPropertyValue',
+        'GetAssetPropertyValueHistory',
+        'ListAccessPolicies',
+        'ListAssets',
+    ],
     'kinesis': ['DescribeStreamConsumer', 'ListShards'],
-    'kinesisvideo': ['DescribeStream', 'ListTagsForStream'],
+    'kinesisvideo': [
+        'DescribeImageGenerationConfiguration',
+        'DescribeNotificationConfiguration',
+        'DescribeSignalingChannel',
+        'DescribeStream',
+        'ListTagsForStream',
+    ],
     'kinesis-video-archived-media': ['GetHLSStreamingSessionURL'],
+    'lightsail': ['GetDistributionLatestCacheReset'],
+    'lookoutmetrics': ['GetSampleData'],
     'mediastore': ['DescribeContainer'],
+    'network-firewall': [
+        'DescribeFirewall',
+        'DescribeFirewallPolicy',
+        'DescribeLoggingConfiguration',
+        'DescribeRuleGroup',
+        'DescribeRuleGroupMetadata',
+    ],
     'opsworks': [
         'DescribeAgentVersions', 'DescribeApps', 'DescribeCommands', 'DescribeDeployments', 'DescribeEcsClusters',
         'DescribeElasticIps', 'DescribeElasticLoadBalancers', 'DescribeInstances', 'DescribeLayers',
         'DescribePermissions', 'DescribeRaidArrays', 'DescribeVolumes'
     ],
+    'personalize-runtime': ['GetRecommendations'],
     'pricing': ['GetProducts'],
-    'redshift': ['DescribeTableRestoreStatus', 'DescribeClusterSecurityGroups'],
+    'redshift': ['DescribeTableRestoreStatus', 'DescribeClusterSecurityGroups', 'DescribeReservedNodeExchangeStatus'],
+    'redshift-serverless': ['GetSnapshot'],
+    'resource-groups': ['GetGroup', 'GetGroupConfiguration', 'GetGroupQuery', 'ListGroupResources'],
+    'robomaker': ['GetWorldTemplateBody'],
     'route53domains': ['GetContactReachabilityStatus'],
     'schemas': ['GetResourcePolicy'],
+    'sagemaker': ['ListAssociations', 'ListPipelineExecutionSteps'],
     'secretsmanager': ['GetRandomPassword'],
     'servicecatalog': [
         'DescribeProduct',
@@ -294,12 +408,19 @@ PARAMETERS_REQUIRED = {
         'DescribeProvisionedProduct',
         'DescribeProvisioningArtifact',
         'DescribeProvisioningParameters',
+        'GetProvisionedProductOutputs',
     ],
     'shield': ['DescribeSubscription', 'DescribeProtection'],
     'sms': ['GetApp', 'GetAppLaunchConfiguration', 'GetAppReplicationConfiguration'],
-    'ssm': ['DescribeAssociation', 'DescribeMaintenanceWindowSchedule', 'ListComplianceItems'],
+    'ssm': [
+        'DescribeAssociation',
+        'DescribeMaintenanceWindowSchedule',
+        'ListComplianceItems',
+        'ListOpsItemRelatedItems',
+    ],
     # TODO: waf ListLoggingConfigurations may just require fixing
     'waf': ['ListActivatedRulesInRuleGroup', 'ListLoggingConfigurations'],
+    'wafv2': ['GetRuleGroup'],
     'waf-regional': ['ListActivatedRulesInRuleGroup'],
     'workdocs': ['DescribeActivities', 'GetResources'],
     # TODO: worklink ListFleets might just require fixing
@@ -310,7 +431,9 @@ PARAMETERS_REQUIRED = {
 
 def get_services():
     """Return a list of all service names where listable resources can be present"""
-    return [service for service in sorted(boto3.Session().get_available_services()) if service not in SERVICE_BLACKLIST]
+    return [
+        service for service in sorted(boto3.Session().get_available_services()) if service not in SERVICE_IGNORE_LIST
+    ]
 
 
 def get_verbs(service):
@@ -381,28 +504,49 @@ def get_endpoint_hosts():
         print('  ...looking for {} in all regions...'.format(service))
         result[service] = {}
         for region in ALL_REGIONS:
-            result[service][region] = boto3.Session(region_name=region).client(service).meta.endpoint_url
-
+            meta = get_client(service, region=region).meta
+            # In some services, different operations must access different host prefixes ("api.", "env.").
+            # This means that the endpoint_url itself may not point to any host, defeating our heuristic.
+            # Therefore, we only pick the base URL if at least one operation accesses it, otherwise we pick the
+            # alphabetically first host prefix.
+            endpoint_prefixes = set(
+                meta.service_model.operation_model(op_name).endpoint.get('hostPrefix')
+                for op_name in meta.service_model.operation_names
+                if meta.service_model.operation_model(op_name).endpoint
+            )
+            if None in endpoint_prefixes or not endpoint_prefixes:
+                result[service][region] = [meta.endpoint_url]
+            else:
+                assert meta.endpoint_url.startswith("https://"), meta.endpoint_url
+                result[service][region] = []
+                prefixes = sorted(endpoint_prefixes)
+                if any(pf.endswith(".") for pf in prefixes):
+                    result[service][region] = [meta.endpoint_url]
+                for prefix in prefixes:
+                    result[service][region].append("https://" + prefix + meta.endpoint_url[len("https://"):])
     print('...done.')
     return result
 
 
-def get_endpoint_ip(service_region_host):
-    (service, region), host = service_region_host
-    try:
-        result = gethostbyname(host.split('/')[2])
+def get_endpoint_ip(service_region_hosts):
+    (service, region), hosts = service_region_hosts
+    result = None
+    for host in hosts:
+        try:
+            result = gethostbyname(host.split('/')[2])
+        except gaierror as ex:
+            if ex.errno != -5:  # -5 is "No address associated with hostname"
+                raise
         return (service, region, result)
-    except Exception:
-        return (service, region, None)
 
 
 def get_service_region_ip_in_dns():
-    service_region_host = {}
-    for service, region_host in get_endpoint_hosts().items():
-        for region, host in region_host.items():
-            service_region_host[(service, region)] = host
+    service_region_hosts = {}
+    for service, region_hosts in get_endpoint_hosts().items():
+        for region, hosts in region_hosts.items():
+            service_region_hosts[(service, region)] = hosts
     print('Resolving endpoint IPs to find active endpoints...')
-    result = ThreadPool(128).map(get_endpoint_ip, service_region_host.items())
+    result = ThreadPool(128).map(get_endpoint_ip, service_region_hosts.items())
     print('...done')
     return result
 
